import numpy as np

class SunPosition:

	def __init__(self):
		""" Calculate sun position according to a certain location, date and time. Reference: John A. Duffie and William A. Beckman. Solar Engineering of Thermal Processes, 4th edition.
		
		``Example``

		  * Location: latitude = 37.44
		  * Date    : 21 Jun
		  * Time    : solar noon or two hours after sunrise
									
			>>> from solsticepy.cal_sun import *
			>>> sun=SunPosition()
			>>> day=sun.days(21, 'Jun')
			>>> latitude=37.44
			>>> delta=sun.declination(day)

			Solar noon

			>>> omega=0.
			>>> theta=sun.zenith(latitude, delta, omega)
			>>> phi=sun.azimuth(latitude, theta, delta, omega)
			>>> print(theta)
			13.987953925483858
			>>> print(phi)
			0.0

			Two hours after sunrise

			>>> daytime, sunrise=sun.solarhour(delta, latitude)
			>>> omega=sunrise+15.*2. # solar noon
			>>> theta=sun.zenith(latitude, delta, omega)
			>>> phi=sun.azimuth(latitude, theta, delta, omega)
			>>> print(theta)
			67.91774797592434
			>>> print(phi)
			-103.31434583806747 

		"""		
		pass


	def days(self, dd, mm):
		"""Calculate the day-of-the-year for specified date and month, ref. J Duffie page 14, Table 1.6.1

		``Arguments``

			* dd (int): the day in the month
			* mm (str): the month, e.g. 'Jan', 'Feb', Mar' etc

		``Return`` 

			* days (int): the-day-of-the year for the specified dd-mm, e.g. 1 Jan is day 1

		"""

		if mm=='Jan':
		    days=dd

		elif mm=='Feb':
		    days=31+dd

		elif mm=='Mar':
		    days=59+dd

		if mm=='Apr':
		    days=90+dd

		elif mm=='May':
		    days=120+dd

		elif mm=='Jun':
		    days=151+dd

		if mm=='Jul':
		    days=181+dd

		elif mm=='Aug':
		    days=212+dd

		elif mm=='Sep':
		    days=243+dd

		if mm=='Oct':
		    days=273+dd

		elif mm=='Nov':
		    days=304+dd

		elif mm=='Dec':
		    days=334+dd

		return days


	def declination(self, days, form='detail'):
		"""Calculate the solar declination angle for a specified day-of-the-year, ref. J Duffie page 13, declination angle: delta=23.45*sin(360*(284+day)/365)

		``Arguments``

		  * day (int): day of the year, i.e from 1 to 365
		  * form (str): 'detail' or simple' model

		``Return``

		  * delta (float): declination angle (deg)
		"""

		if form=='detail':
		    B=float(days-1)*360./365.*np.pi/180.

		    delta=(180./np.pi)*(0.006918 - 0.399912*np.cos(B) +0.070257*np.sin(B)- 0.006758*np.cos(2.*B) + 0.000907*np.sin(2.*B)- 0.002697*np.cos(3.*B) + 0.00148*np.sin(3.*B))

		else:
		    delta=23.45*np.sin(360.*float(284+days)/365.*np.pi/180.) # deg

		return delta


	def solarhour(self, delta, latitude):
		"""Calculate day length and sunrise hour-angle, ref. J Duffie page 17

		``Arguments``

		  * delta (float): declination angle (deg)
		  * latitude (float): latitude angle (deg)

		``Returns``

		  * hour (float): length of the daylight hour (h)
		  * sunrise (float): the solar hour angle at sunrise (deg)

		"""

		sunset=np.arccos(-np.tan(latitude*np.pi/180.)*np.tan(delta*np.pi/180.))*180./np.pi # deg

		sunrise=-sunset

		hour=(sunset-sunrise)/15.

		return hour, sunrise


	def zenith(self, latitude, delta, omega):
		"""Calculate the zenith angle, ref. J Duffie eq.1.6.5

		``Arguments``

		  * latitude (float): latitude angle at the location (deg)
		  * delta (float):  declination angle (deg)
		  * omega (float): solar hour angle (deg)

		``Return``

		  * theta (float): the zenith angle (deg)

		"""            
		latitude*=np.pi/180.
		delta*=np.pi/180.
		omega*=np.pi/180.

		theta=np.arccos(np.cos(latitude)*np.cos(delta)*np.cos(omega)+np.sin(latitude)*np.sin(delta))*180./np.pi

		return theta
		
	def azimuth(self, latitude, theta, delta, omega):
		"""Calculate azimuth angle at specified location and sun position, ref. J Duffie eq. 1.6.6

		``Arguments``

		  * latitude (float): latitude angle (deg)
		  * delta (float): declination angle (deg)
		  * theta (float): zenith angle (deg)
		  * omega (float): solar hour angle (deg)

		``Return``
 
		  * phi (float): azimuth angle (deg), counted from South towards to West 
		"""
		latitude*=np.pi/180.
		delta*=np.pi/180.
		theta*=np.pi/180.

		a1=np.cos(theta)*np.sin(latitude)-np.sin(delta)
		a2=np.sin(theta)*np.cos(latitude)
		b=a1/a2

		if abs(b+1.)<1e-10:
		    phi=np.pi

		elif abs(b-1.)<1e-10:
		    phi=0.
		else:
		    phi=abs(np.arccos((np.cos(theta)*np.sin(latitude)-np.sin(delta))/(np.sin(theta)*np.cos(latitude)))) # unit radian

		if omega<0:
		    phi=-phi

		phi*=180./np.pi

		return phi

	def convert_AZEL_to_declination_hour(self, theta, phi, latitude):
		""" Convert azimuth-elevation angle to declination-hour angle

		``Arguments``

		  * theta (float): zenith angle (deg)
		  * phi (float): azimuth angle (deg), counted from South towards to West
		  * latitude (float): latitude latitude (deg)

		``Returns``

		  * delta: declination angle (deg)
		  * omega: solar hour angle (deg)
		"""
     
		phi*=np.pi/180.
		theta*=np.pi/180.
		latitude*=np.pi/180.

		delta=np.arcsin(np.cos(theta)*np.sin(latitude)-np.cos(abs(phi))*np.sin(theta)*np.cos(latitude))

		omega=np.arccos((np.cos(theta)-np.sin(latitude)*np.sin(delta))/(np.cos(latitude)*np.cos(delta)))
		if phi<0:
		    omega=-omega

		delta*=180./np.pi
		omega*=180./np.pi

		return delta, omega

	def convert_convention(self, tool, azimuth, zenith):
		"""Return azimuth-elevation angle using the angle convention of Solstice or SolarTherm

		``Arguments``

		  * tool (str): 'solstice' or 'solartherm'
		  * azimuth (float): azimuth angle (deg), counted from South towards to West
		  * zenith (float): zenith angle (deg)


		``Returns``

		  * sol_azi (float): azimuth angle 
		  * sol_ele (float): elevation angle
		"""
		if tool=='solstice':
		    # azimuth: from East to North 
		    sol_azi=-(90.+azimuth)
		    sol_ele=90.-zenith 

		elif tool=='solartherm':
		    # azimuth: from North to Ease (clockwise)
		    # ref Ali email (20/3/18)
		    sol_azi=180.+azimuth
		    sol_ele=90.-zenith  

		if isinstance(sol_azi, np.ndarray):
		    for i in range(len(sol_azi)):
		        azi=sol_azi[i]
		        ele=sol_ele[i]
		        if (azi>=360. or azi<0.):
		            sol_azi[i]=(azi+360.)%360.
		        if ele<=1e-20:
		            sol_ele[i]=0.  
		else:
		    if (sol_azi>=360. or sol_azi<0.):
		        sol_azi=(sol_azi+360.)%360.
		    if sol_ele<=1e-20:
		        sol_ele=0.  
		
		return sol_azi, sol_ele        
		                   

	def annual_angles(self, latitude, casefolder='NOTSAVE', nd=5, nh=5):
		"""Generate a range of sun positions (azimuth-zenith angles and declination-solarhour angles) for annual optical lookup table simulation. Automatically detect the time when the sun is below the horizon (elevation<0), where a ray-tracing simulation is not required.
		
		``Arguments``

		  * latitude (float): latitude of the location (deg)
		  * casefolder (str): directory to save the table and case_list in .csv files, or 'NOTSAVE' (by default) to not write the output to files
		  * nd (int): number of rows of the lookup table (points in the declination movement, suggest nd>=5)
		  * nh (int): number of columns of the lookup table (hours in a day, i.e. 24h)

		``Returns``

		  * AZI (1 D numpy array): a list of azimuth angles (deg), counted from South towards to West
		  * ZENITH (1D numpy array): a list of zenith angles (deg)
		  * table (numpy array): declination (row) - solarhour (column) lookup table to be simulated 
		  * case_list (numpy array): a flatten list of cases to be simulated, with the correspondence between declination-solarhour angles and azimuth-zenith angles for each case


		``Example``

			>>> from solsticepy.cal_sun import *
			>>> sun=SunPosition()
			>>> latitude=37.44
			>>> casefolder='.'
			>>> nd=5
			>>> nh=9
			>>> sun.annual_angles(latitude, casefolder, nd, nh)

			A 5x9 lookup table will be generated to the current directory:

			+--------------+----+-------+-------+--------+--------+--------+--------+--------+--------+--------+--------+
			| Lookup table |    |       |                      Solar hour angles (deg)                                  | 
			+==============+====+=======+=======+========+========+========+========+========+========+========+========+
			|              |    |       |   1   |    2   |   3    |    4   |    5   |    6   |   7    |    8   |    9   |
			+              +----+-------+-------+--------+--------+--------+--------+--------+--------+--------+--------+
			|              |    |       | -180  |  -135  |  -90   |  -45   |    0   |   45   |   90   |   135  |  180   |
			+              +----+-------+-------+--------+--------+--------+--------+--------+--------+--------+--------+
			|              | 1  |-23.45 |  `-`  |   `-`  |   `-`  | case 1 | case 2 | ***1   |   `-`  |   `-`  |   `-`  |
			+ Declination  +----+-------+-------+--------+--------+--------+--------+--------+--------+--------+--------+
			|              | 2  |-11.73 |  `-`  |   `-`  |   `-`  | case 3 | case 4 | ***3   |   `-`  |   `-`  |   `-`  |
			+ angles       +----+-------+-------+--------+--------+--------+--------+--------+--------+--------+--------+
			|              | 3  |   0   |  `-`  |   `-`  | case 5 | case 6 | case 7 | ***6   | ***5   |   `-`  |   `-`  |
			+ (deg)        +----+-------+-------+--------+--------+--------+--------+--------+--------+--------+--------+
			|              | 4  | 11.73 |  `-`  |   `-`  | case 8 | case 9 | case 10| ***8   | ***9   |   `-`  |   `-`  |
			+              +----+-------+-------+--------+--------+--------+--------+--------+--------+--------+--------+
			|              | 5  | 23.45 |  `-`  |   `-`  | case 11| case 12| case 13| ***11  | ***12  |   `-`  |   `-`  |
			+--------------+----+-------+-------+--------+--------+--------+--------+--------+--------+--------+--------+

			  * case `n`: this sun position is the `n`-th case to be simulated
			  * ***`n`  : it is the symmetric case with the case `n`, no re-simulation is required
			  * `-`     : the sun is below the horizon, no simulation is required
		"""

		# declination angle (deg)  
		# -23.45 ~ 23.45
		DELTA=np.linspace(-23.45, 23.45, nd)
 
		# solar time
		solartime=np.linspace(-180., 180., nh) # deg

		table=np.zeros(((nh+3)*(nd+3)))
		table=table.astype(str)
		for i in range(len(table)):
			table[i]=''

		table=table.reshape(nd+3, nh+3)
		table[0,0]='Lookup table'
		table[3,0]='Declination (deg)'
		table[0,3]='Hour angle (deg)'

		table[3:,1 ]=np.arange(1,nd+1)
		table[1 ,3:]=np.arange(1,nh+1)

		table[3:,2 ]=DELTA   
		table[2 ,3:]=solartime

		c=1
		AZI=np.array([])
		ZENITH=np.array([])

		case_list=np.array(['Case','declination (deg)','solar hour angle (deg)', 'azimuth (deg) S-to-W ', 'zenith (deg)'])

		for i in range(nd):
			delta=DELTA[i]
			hour, sunrise=self.solarhour(delta, latitude)
			sunset=-sunrise
			for j in range(nh):
				omega=solartime[j]
				if (omega>sunset or omega<sunrise):
					table[3+i,3+j]='-' 

				else:
					if omega<0:

						table[3+i, 3+j]=' case %s'%(c)
						table[3+i, -(1+j)]='***%s'%(c)

						#zenith angle
						theta=self.zenith(latitude, delta, omega)
						# azimuth        
						phi=self.azimuth(latitude, theta, delta, omega)
						AZI=np.append(AZI, phi)
						ZENITH=np.append(ZENITH, theta)

						case_list=np.append(case_list, (c, delta, omega, phi, theta)) 

						#zenith angle (afternoon)
						theta=self.zenith(latitude, delta, -omega)
						# azimuth        
						phi=self.azimuth(latitude, theta, delta, -omega)    
						case_list=np.append(case_list, (c, delta, -omega, phi, theta))                 

						c+=1

					elif omega==0:
						table[3+i, 3+j]=' case %s'%(c)

						#zenith angle
						theta=self.zenith(latitude, delta, omega)
						# azimuth        
						phi=self.azimuth(latitude, theta, delta, omega)
						AZI=np.append(AZI, phi)
						ZENITH=np.append(ZENITH, theta)

						case_list=np.append(case_list, (c, delta, omega, phi, theta)) 
						c+=1

		              	                                     
		case_list=case_list.reshape(int(len(case_list)/5),5)
		#azimuth=case_list[1:,-2].astype(float)
		#zenith=case_list[1:,-1].astype(float)

		if casefolder!='NOTSAVE':    
		    np.savetxt(casefolder+'/table_view.csv', table, fmt='%s', delimiter=',')  
		    np.savetxt(casefolder+'/annual_simulation_list.csv', case_list, fmt='%s', delimiter=',')          

		return AZI, ZENITH, table,case_list



if __name__=='__main__':
    # example: PS10, summer solstice, solar noon
    latitude=37.44

    sun=SunPosition()
    dd=sun.days(21, 'Jun')
    delta=sun.declination(dd)
    print('Declination angle', delta)


    daytime,sunrise=sun.solarhour(delta, latitude)
    print('Day timeS', daytime)
    print('sun rise', sunrise)
    print('')

    omega=0. # solar noon
    theta=sun.zenith(latitude, delta, omega)
    phi=sun.azimuth(latitude, theta, delta, omega)
    print('elevation', 90.-theta)
    print('azimuth', phi)

    azi, zen, table,caselist=sun.annual_angles(latitude, hemisphere='North', casefolder='.',nd=5, nh=7)


    
    



        
        
        



import numpy as np
from datetime import datetime

def output_motab(table,savedir=None, title=None):
	'''
	output the .motab table fiel
	'''
	f=open(savedir, 'w')
	f.write('#1\n')

	if table.ndim==2:
		# size of the lookup table
		m=np.shape(table)[0]-2
		n=np.shape(table)[1]-2
		f.write('double optics(%s, %s)\n'%(m,n))

		hour_angle=table[2, 3:]
		declination=table[3:,2]

		for i in range(m):
			if i ==0:
				row_i=np.append(0, hour_angle)
			else:
				row_i=np.append(declination[i-1], table[2+i, 3:])

			#content=np.array2string(row_i, formatter={'float_kind':lambda x: "%.2f" % row_i})
			#content=np.array2string(row_i, precision=2, separator=' ', suppress_small=True)
			#f.write(content+'\n')
			f.write(" ".join(map(str, row_i)))
			f.write("\n")

	else:
		# 3D table, include the breakdown of the total energy
		a=len(table)
		m=np.shape(table[0])[0]-2
		n=np.shape(table[0])[1]-2

		hour_angle=table[0][2, 3:]
		declination=table[0][3:,2]

		for t in range(a):
			f.write('double %s(%s, %s)\n'%(title[t], m,n))

			for i in range(m):
				if i ==0:
					row_i=np.append(0, hour_angle)
				else:
					row_i=np.append(declination[i-1], table[t][2+i, 3:])
				f.write(" ".join(map(str, row_i)))
				f.write("\n")	
						
			f.write("\n")
			f.write("\n")

	f.close()


def output_matadata_motab(table, field_type, aiming, n_helios, A_helio, eff_design, H_rcv, W_rcv, H_tower, Q_in_rcv, A_land, savedir=None, details_en=None):
	"""Output the .motab file to work with the SolarTherm program

	``Arguments``
		* table (numpy array): the oelt that returned by design_crs.CRS.field_design_annual, listed by declination and hour angles
		* field_type (str): polar or surrounding
		* aiming (str): 'single' or 'isp' or others to specify the aiming strategy
		* n_helios (int): total number of heliostats
		* A_helio (float): area of each heliostat (m2)
		* eff_design (float): the optical efficiency at design point
		* H_rcv (float): height of the heliostat (m)
		* W_rcv (float): width of the heliostat (m)
		* H_tower (float), tower height (m)
		* Q_in_rcv (float): the incident power on the receiver (W)
		* A_land (float): the total land area of the field (m2)
		* savedir (str): the directory to save this .motab file
		* details_en (dict): the key of the dict is the name of the breakdown of energy (loss), the value of the dict is a numpy array that contains the value of the energy (loss) with the corresponding declination and hour angles
	
	``Return``
		write the table(s) to the .motab file
	"""
	f=open(savedir, 'w')
	f.write('#1\n')
	f.write('#Comments: Field type: %s, Aiming Strategy: %s, Date:%s\n'%(field_type, aiming, datetime.now()))
	f.write('#METALABELS,n_helios,A_helio,Eff_design,H_rcv,W_rcv,H_tower, Q_in_rcv, A_land\n')
	f.write('##METAUNITS,real,m2,real,m,m,m,W,m2\n')
	f.write('#METADATA,%s,%s,%s,%s,%s,%s,%s,%s\n'%(n_helios,A_helio,eff_design,H_rcv,W_rcv,H_tower,Q_in_rcv,A_land))

	# size of the lookup table  
	m=np.shape(table)[0]-2
	n=np.shape(table)[1]-2
	f.write('double optics(%s, %s)\n'%(m,n))

	hour_angle=table[2, 3:]
	declination=table[3:,2]

	for i in range(m):
		if i ==0:
			row_i=np.append(0, hour_angle)
		else:
			row_i=np.append(declination[i-1], table[2+i, 3:])

		#content=np.array2string(row_i, formatter={'float_kind':lambda x: "%.2f" % row_i})
		#content=np.array2string(row_i, precision=2, separator=' ', suppress_small=True)
		#f.write(content+'\n')
		f.write(" ".join(map(str, row_i)))
		f.write("\n")

	if details_en!=None:
		for key in details_en:
			breakdown=key
			table=details_en[key]
			# size of the lookup table  
			m=np.shape(table)[0]-2
			n=np.shape(table)[1]-2
			f.write('double %s(%s, %s)\n'%(key,m,n))
			for i in range(m):
				if i ==0:
					row_i=np.append(0, hour_angle)
				else:
					row_i=np.append(declination[i-1], table[2+i, 3:])
			f.write(" ".join(map(str, row_i)))
			f.write("\n")				
	f.close()


def read_motab(filename):

	with open(filename) as f:
		content=f.read().splitlines()
	f.close()
	res=content[4].split(',')

	n_helios=float(res[1])
	A_helio=float(res[2])
	eff_des=float(res[3])
	Q_in_rcv=float(res[-2])
	A_land=float(res[-1])

	oelt=np.array([])
	solar_hour=np.array([])
	declination=np.array([])
	t=content[6].split(' ')

	for v in t[1:]:
		solar_hour=np.append(solar_hour, float(v))

	for t in content[7:]:
		v=t.split(' ')
		declination=np.append(declination, float(v[0]))
		oelt=np.append(oelt, np.array(v[1:],dtype=float))

	oelt=oelt.reshape(len(declination), len(solar_hour))

	return n_helios, A_helio, eff_des, Q_in_rcv, A_land, solar_hour, declination, oelt
	





import numpy as np
import sys
import matplotlib
matplotlib.use("agg")
import matplotlib.pyplot as plt
from .cal_sun import *
from .gen_vtk import gen_vtk

def radial_stagger(latitude, num_hst, width, height, hst_z, towerheight, R1, fb, dsep=0., field='polar', savedir='.', plot=False):
	'''Generate a radial-stagger heliostat field, ref. Collado and Guallar, 2012, Campo: Generation of regular heliostat field.

	``Arguments``
	  * latitude (float): latitude of the field location (deg)
	  * num_hst (int)   : number of heliostats
	  * width (float)   : mirror width (m)
	  * height (float)  : mirror height (m)
	  * hst_z (float)   : the vertical location of each heliostat (m)
	  * towerheight (float): tower height (m)
	  * R1 (float)      : distance from the first row to the bottom of the tower, i.e. (0, 0, 0)
	  * fb (float)      : the field layout growing factor, in (0, 1)
	  * dsep (float)    : separation distance (m)
	  * field (str)     : 'polar-half' or 'surround-half' or 'polar' or 'surround' field, the 'half' option is for simulation a symmetric field
	  * savedir (str)   : directory of saving the pos_and_aiming.csv
	  * plot (bool)     : True - plot the layout by Matplotlib

	``Returns``

	  * pos_and_aiming (nx7 numpy array): position, focal length and aiming point of each generated heliostat
	  * a pos_and_aiming.csv file is created and written to the savedir


	``Example``

		>>> from solsticepy.cal_layout import *
		>>> latitude=34.
		>>> hst_width=13.
		>>> hst_height=10.
		>>> hst_z=5.
		>>> target_area=75715.
		>>> num_hst=target_area/hst_width/hst_height*2 # create a twice large field, so that the inefficient heliostat can be trimmed-off after optical simulations
		>>> tower_height=110.
		>>> R1=40. # the distance from the first row to the bottom of the tower
		>>> fb=0.6
		>>> casefolder='.' # replace it as the directory of your case folder
		>>> plot=False # if you want to plot the field layout, set it as True
		>>> pos_and_aim=radial_stagger(latitude, num_hst, hst_width, hst_height, hst_z, tower_height, R1, fb, savedir=casefolder,plot=plot)

		An nx7 array is returned, and the pos_and_aim.csv file is saved in the local directory

	'''

	# heliostat diagonal distantce
	DH=np.sqrt(height**2+width**2) 

	# distance between contiguous helistat center on the X and Y plane
	DM=DH+dsep

	# minimum radial increment
	delta_Rmin=0.866*DM

	# number of heliostats in the first row
	Nhel1 =int(2.*np.pi*R1/DM)


	# the total number of zones (estimated)
	#Nzones=int(np.log(5.44*3*(num_hst/az_rim*np.pi)/Nhel1**2+1)/np.log(4))+1

	X={}
	Y={}
	Nrows_zone=np.array([])
	Nhel_zone=np.array([])
	delta_az_zone=np.array([])

	num=0
	i=0
	sys.stderr.write('DM '+repr(DM)+'\n')
	sys.stderr.write('dRm'+repr(delta_Rmin)+'\n')

	while num<num_hst*3:
		Nrows= int((2.**(i))*Nhel1/5.44)
		Nhel=(2**(i))*Nhel1
		R=Nhel/2./np.pi*DM
		delta_az=2.*np.pi/Nhel

		Nrows_zone=np.append(Nrows_zone, Nrows)
		Nhel_zone=np.append(Nhel_zone, Nhel)
		delta_az_zone=np.append(delta_az_zone, delta_az)

		nh=np.arange(Nhel)
		azimuth=np.zeros((int(Nrows), int(Nhel)))
		azimuth[0::2, :]=delta_az/2.+nh*delta_az-np.pi/2. # the odd rows
		azimuth[1::2, :]=nh*delta_az-np.pi/2.

		row=np.arange(Nrows)
		r=R+row*delta_Rmin

		xx=r[:, None]*np.sin(azimuth)
		yy=r[:, None]*np.cos(azimuth)

		X[i]=xx
		Y[i]=yy
		num+=len(xx.flatten())
		i+=1
	Nzones=i

	sys.stderr.write("\n")
	sys.stderr.write("Denest field %d\n"%(num))

	if field=='surround':
		num_hst*=2

	# expanding the field
	#
	wr=width/height
	const=(1.-(1.-fb)*wr/(2.*wr-(np.sqrt(1.+wr**2)+dsep/height)))*height

	XX=np.array([])
	YY=np.array([])
	ZONE=np.array([])  # zone index
	ROW=np.array([])   # row index among the rows in a zone
	TTROW=np.array([]) # row index among the total rows
	NHEL=np.array([])  # No. index among the heliostats in a row

	for i in range(Nzones):
		Nrows=int(Nrows_zone[i])
		Nhel=int(Nhel_zone[i])
		delta_az=delta_az_zone[i]

		R=np.zeros((Nrows, Nhel))

		if i==0:
			# first zone
			R[0]=R1 # first row
		else:
			# second zones
			R[0,  ::2]=Rn+1.5*DRn
			R[0,  1::2]=Rn+1.5*DRn            
			#R[0,-1]=0.5*(R[0,0]+Rn[-1])

		xx=X[i]
		yy=Y[i]
		zz=np.ones(np.shape(xx))*hst_z
		cosw, coseT=cal_cosw_coset(latitude, towerheight,xx, yy, zz)
		row=np.arange(Nrows)
		cosw=cosw.reshape(Nrows, Nhel)  
		coseT=coseT.reshape(Nrows, Nhel)

		Delta_R=cosw/coseT*const
		Delta_R[Delta_R<delta_Rmin]=delta_Rmin

		for j in range(1, Nrows):
			R[j]=R[j-1]+Delta_R[j-1]

		Rn=R[-1]
		DRn=Delta_R[-1]

		nh=np.arange(Nhel)
		azimuth=np.zeros((Nrows, Nhel))
		azimuth[0::2, :]=delta_az/2.+nh*delta_az-np.pi/2. # the odd rows
		azimuth[1::2, :]=nh*delta_az-np.pi/2.

		azimuth=azimuth.flatten()
		R=R.flatten()
		if field=='polar':
			if i<2:
				idx= (azimuth>-np.pi/2.)*(azimuth<np.pi/2.)

			else:
				idx=(azimuth>-np.pi/2.+i*np.pi/40.)* (azimuth<np.pi/2.-i*np.pi/40.)

			xx=R[idx]*np.sin(azimuth[idx])
			yy=R[idx]*np.cos(azimuth[idx])

		else:                       
			xx=R*np.sin(azimuth)
			yy=R*np.cos(azimuth)       
		XX=np.append(XX, xx)
		YY=np.append(YY, yy)
		
		nhels, rows=np.meshgrid(nh, row)
		zone=np.ones(np.shape(rows))*i
		ZONE=np.append(ZONE, zone)
		ROW=np.append(ROW, rows)
		NHEL=np.append(NHEL, nhels)
		if len(TTROW)==0:
			TTROW=np.append(TTROW, rows)
		else:			
			TTROW=np.append(TTROW, rows+np.max(TTROW)+1)
					
	num_hst=int(num_hst)
	XX=XX[:num_hst]
	YY=YY[:num_hst]
	ZONE=ZONE[:num_hst]
	ROW=ROW[:num_hst]
	NHEL=NHEL[:num_hst]
	TTROW=TTROW[:num_hst]

	sys.stderr.write("\nExpanded field %d\n"%(num_hst,))

	hstpos=np.zeros(num_hst*3).reshape(num_hst, 3)
	hstpos[:, 0]=XX
	hstpos[:, 1]=YY
	hstpos[:,2]=hst_z

	# the aiming point is a default point
	# it is required to be revised for receiver performance
	aim_x=np.zeros(num_hst)
	aim_y=np.zeros(num_hst)
	aim_z=np.ones(num_hst)*towerheight

	foc=np.sqrt((XX-aim_x)**2+(YY-aim_y)**2+(hstpos[:,2]-aim_z)**2)

	pos_and_aiming=np.append(XX, (YY, hstpos[:,2], foc, aim_x, aim_y, aim_z, ZONE, ROW, NHEL, TTROW))
	title=np.array(['x', 'y', 'z', 'foc', 'aim x', 'aim y', 'aim z', 'Zone', 'Row', 'No.', 'row index',  'm', 'm', 'm', 'm', 'm', 'm', 'm', '-', '-', '-', '-'])
	pos_and_aiming=pos_and_aiming.reshape(11,int(len(pos_and_aiming)/11))
	pos_and_aiming=np.append(title, pos_and_aiming.T)
	pos_and_aiming=pos_and_aiming.reshape(int(len(pos_and_aiming)/11), 11)

	np.savetxt('%s/pos_and_aiming.csv'%savedir, pos_and_aiming, fmt='%s', delimiter=',')

	if plot:
		fts=24
		plt.figure(dpi=100.,figsize=(12,9))
		plt.plot(XX, YY, '.')
		#plt.xlim(-1000, 1000)
		#plt.ylim(-1000, 1000)
		plt.xticks(fontsize=fts)
		plt.yticks(fontsize=fts)
		plt.xlabel('x (m)', fontsize=fts)
		plt.ylabel('y (m)', fontsize=fts)
		plt.savefig(savedir+'/field_layout.png', bbox_inches='tight')
		plt.close()

	return pos_and_aiming

def cal_cosw_coset(latitude, towerheight, xx, yy, zz):
	'''
	The factors to growing the heliostat field, see eq.(2) Francisco J. Collado, Jesus Guallar, Campo: Generation of regular heliostat fields, 2012

	``Arguments``
	  * latitude (float)   : latitude of the field location (deg)
	  *	towerheight (float): tower height (m)
      * xx, yy, zz (float) : coordinates of heliostats

	``Returns``
	  * cosw (array) : cos(omega)
	  * coseT (array): cos(epsilon_T) 

	'''

	hst_pos=np.append(xx, (yy, zz))
	hst_pos=hst_pos.reshape(3, len(xx.flatten())) # 3 x n
	tower_vec=-hst_pos
	tower_vec[-1]+=towerheight
	tower_vec/=np.sqrt(np.sum(tower_vec**2, axis=0)) # 3 x n
	unit=np.array([[0.], [0.], [1.]])
	unit=np.repeat(unit, len(tower_vec[0]), axis=1)
	coseT=np.sum(tower_vec*unit, axis=0)

	sun=SunPosition()
	dd=sun.days(21, 'Mar')
	delta=sun.declination(dd)
	h=np.arange(8, 17)

	latitude=latitude/180.*np.pi
	delta=delta/180.*np.pi
	omega=(-180.+15.*h)/180.*np.pi
	theta=np.arccos(np.cos(latitude)*np.cos(delta)*np.cos(omega)+np.sin(latitude)*np.sin(delta))

	a1=np.cos(theta)*np.sin(latitude)-np.sin(delta)
	a2=np.sin(theta)*np.cos(latitude)
	b=a1/a2
	phi=abs(np.arccos((np.cos(theta)*np.sin(latitude)-np.sin(delta))/(np.sin(theta)*np.cos(latitude)))) # unit radian
	phi[abs(b+1.)<1e-10]=np.pi
	phi[abs(b-1.)<1e-10]=0.
	phi[omega<0]=-phi[omega<0]
	 
	cosw=np.zeros(len(tower_vec[0]))
	sun_z = np.cos(theta)
	sun_y=-np.sin(theta)*np.cos(phi)
	sun_x=-np.sin(theta)*np.sin(phi)
	sun_vec = np.append(sun_x, (sun_y, sun_z))
	sun_vec=sun_vec.reshape(3, len(sun_x)) # 3xs

	for i in range(len(sun_vec[0])):
		sv=np.repeat(sun_vec[:,i], len(tower_vec[0])).reshape(3, len(tower_vec[0]))
		hst_norms=sv+tower_vec
		hst_norms/=np.linalg.norm(hst_norms, axis=0)
		cosw+=np.sum(sv*hst_norms, axis=0)
	cosw/=float(len(sun_vec[0]))
	return cosw, coseT


def aiming_cylinder(r_height,r_diameter, pos_and_aiming, savefolder, c_aiming=0.):
	'''
	The aiming method is developed following the deviation-based multiple aiming, by Shuang Wang. Reference: Augsburger G. Thermo-economic optimisation of large solar tower power plants[R]. EPFL, 2013.

	``Arguments``
	  * r_height (float)   : receiver height (m)
	  * r_diameter (float) : receiver diameter (m)
	  * pos_and_aiming (array): the array returned from the function radial_stagger
	  * savefolder (str)   : directory to save the aiming point results
	  * c_aiming (float)   : an aiming co-efficient

	``Returns``

	  * hst_info_ranked (array) : the heliostats ranked according to focal lenghts	
	  * a pos_and_aiming.csv file is created and written to the savedir, which contains pos_and_aiming (nx7 numpy array): position, focal length and the updated aiming point of each heliostat 

	'''
	r_radius=0.5*r_diameter
	hst_info=pos_and_aiming[2:].astype(float)
	num_hst=hst_info.size/7
	#print num_hst
	foc=hst_info[:,3] # the focal lenghts of all the heliostats

	# ranked the hsts according to focal lenghts
	hst_info_ranked = hst_info[np.argsort(foc)[::1]]

	for i in range(num_hst):
		
		if (i+1)%2==0: # negative
			hst_info_ranked[i,6]=hst_info_ranked[i,6]+0.5*r_height*c_aiming*(float(num_hst)-1-i)/num_hst
		else:
			hst_info_ranked[i,6]=hst_info_ranked[i,6]-0.5*r_height*c_aiming*(float(num_hst)-1-i)/num_hst
		
		hst_info_ranked[i,4]=hst_info_ranked[i,0]*r_radius/np.sqrt(hst_info_ranked[i,0]**2+hst_info_ranked[i,1]**2)
		hst_info_ranked[i,5]=hst_info_ranked[i,1]*r_radius/np.sqrt(hst_info_ranked[i,0]**2+hst_info_ranked[i,1]**2)
		#print hst_info_ranked[i,0],hst_info_ranked[i,1],hst_info_ranked[i,4],hst_info_ranked[i,5]
		hst_info_ranked[i,3]=np.sqrt((hst_info_ranked[i,0]-hst_info_ranked[i,4])**2+(hst_info_ranked[i,1]-hst_info_ranked[i,5])**2+(hst_info_ranked[i,2]-hst_info_ranked[i,6])**2)


	title=np.array(['x', 'y', 'z', 'foc', 'aim x', 'aim y', 'aim z', 'm', 'm', 'm', 'm', 'm', 'm', 'm'])

	pos_and_aiming_new=np.append(title, hst_info_ranked)
	pos_and_aiming_new=pos_and_aiming_new.reshape(len(pos_and_aiming_new)/7, 7)

	csv_new=savefolder+'/pos_and_aiming.csv'# the output field file
	np.savetxt(csv_new, pos_and_aiming_new, fmt='%s', delimiter=',')

	return hst_info_ranked	



if __name__=='__main__':
    
	pos_and_aim=radial_stagger(latitude=34., num_hst=22640, width=10., height=10., hst_z=0., towerheight=250, R1=80, fb=0., dsep=0., field='polar', savedir='.', plot=True)

        



from __future__ import print_function
import numpy as np

#for python 2:
#from builtins import super

from .data_spectral import SolarSpectrum, MirrorRhoSpectrum
import sys

class Sun:
	"""Sun parameters for solstice-input

	Example:
	>>> sun = Sun(dni=1000, sunshape='buie', csr=0.2)
	>>> print(sun.yaml(spectrum = "*solar_spectrum"))

	"""
	def __init__(self,dni=1000,sunshape=None,csr=0.01,half_angle_deg=0.2664,std_dev=0.2):
		"""Define sun parameters for Solstice input file.

		`dni`: Direct normal irradance (W/m2)
		`sunshape`: Sunshape: can be None, ``'pillbox'``,``'gaussian'`` or ``'buie'``
		`half_angle_deg`: sun angular size (in DEGREES, half-angle) (ONLY in case of ``'pillbox'``)
		`csr`: circumsolar ratio (ONLY in case of ``'buie'``)
		`std_dev`: standard deviation of the angular dsn ratio (ONLY in case of ``'gaussian'``)
		"""
		self.dni = dni
		self.sunshape = sunshape
		if sunshape is not None:
			assert sunshape in ['buie','pillbox','gaussian']
			self.sunshape = sunshape
			if sunshape == "pillbox":
				self.half_angle_deg = half_angle_deg
			elif sunshape == "buie":
				self.csr = csr
			elif sunshape == "gaussian":
				self.std_dev = std_dev
	def yaml(self,spectrum = None):
		"""YAML representation of the sun for solstice-input.

		`spectrum`: YAML value of the solar spectrum. Would normally be set to the string ``"*solar_spectrum"``.
		"""
		# FIXME surely we find a smarter way to do this serialization AND deseralization with YAML?
		s = "{dni: %15.8e" % (self.dni,)
		if spectrum is not None:
			s += ", spectrum = %s" % (spectrum,)
		if self.sunshape is not None:
			if self.sunshape=='pillbox':
				s += ", pillbox: {half_angle: %6.4f}" % (self.half_angle_deg,)   
			elif self.sunshape=='buie':
				s += ", buie: {csr: %6.4f}" % (self.csr,) 
			elif self.sunshape=='gaussian':
				s += ", gaussian: {std_dev: %6.4f}" % (self.std_dev,)
		s += "}"
		return s


def gen_yaml(sun, hst_pos, hst_foc, hst_aims,hst_w, hst_h
		, rho_refl, slope_error, receiver, rec_param, rec_abs
		, outfile_yaml, outfile_recv
		, hemisphere='North', tower_h=0.01, tower_r=0.01,  spectral=False
		, medium=0, one_heliostat=False
):
	"""Generate the heliostat field and receiver YAML input files for Solstice ray-tracing simulation.

	1. the sun
	  * `sun` (`Sun` object): parameters relating to the solar source

	2. the field
	  * `hst_pos` (nx3 numpy array): heliostat positions (x, y, z) (first of the 'field' parameters)
	  * `hst_foc` (nx1 numpy array): heliostat focal length
	  * `hst_aims` (nx3 numpy array): heliostat aiming point (ax, ay, az)
	  * `hst_w` (float): heliostat mirror width  (in x direction)
	  * `hst_h` (float): heliostat mirror height (in y direction)
	  * `hst_z` (float): heliostat center height (in z direction)
	  * `rho_refl` (float): reflector reflectivity
	  * `slope_error` (float): reflector surface slope error rms, radians
	  * `tower_h` (float): tower height (m)
	  * `tower_r` (float): tower radius (a cylindrical shape tower) (m)
	3. the receiver
	  * `receiver` (str): ``'flat'``, ``'cylinder'``, or ``'stl' or 'multi_cavity'`` (first of the 'receiver' parameters)
	  * `rec_abs` (float): receiver absorptivity
	  * `rec_param` (numpy array or str): each element contains the geometrical parameter of the corresponding receiver.
	4. others
	  * `spectral` (bool): True - simulate the spectral dependent performance (first of the 'other' parameters)
	  * `medium` (float): if the atmosphere is surrounded by non-participant medium, medium=0; otherwise it is the extinction coefficient in m-1
	  * `one_heliosat` (boolean): if `True`, implements ray tracing from just one heliostat.
	  	
	Returns: nothing (requested files are created and written)

	Note that the parameters are in groups that relate to the `sun`, the `field` and the `receiver` then `others`. 

	Note also the type for `rec_param` should be as follows.
	  * if ``receiver == 'flat'``: np.array([width, height, slices, x, y, z, tilt angle (deg))]
	  * if ``receiver == 'cylinder'``: np.array([radius, height, slices])
	  * if ``receiver == 'stl'``: the directory of the stl file
	  * if ``receiver == 'multi_cavity'``: the directory of the stl file
	"""
	# FIXME Parameters should be named according to what they are, eg
	# the parameter should be called 'csr', not 'sunsize', to avoid confusion.
	# We can still improve our calling convention, to make this library easier
	# to use and more maintainable.

	sys.stderr.write("Generating YAML file...\n")

	iyaml='' # the input yaml file

	# 
	### Section (1)
	# set the spectral data: 
	# solar radiative intensity, refractive indexes, extinction coefficients, reflectivities
	#------------------------------
	if spectral:
		I_sun=SolarSpectrum()
		# CREATE the spectrum for the sun
		iyaml+='- spectrum: &solar_spectrum  \n'
		for i in range(0,len(I_sun)-1):
		    iyaml+='  - {wavelength: %s, data: %s }\n' % (I_sun[i][0],I_sun[i][1])  
		i = len(I_sun)-1
		iyaml+='  - {wavelength: %s, data: %s }\n' % (I_sun[i][0],I_sun[i][1]) 
		iyaml+='\n'

		# CREATE the spectrum for the reflectivity (mirror)
		mirror_rho= MirrorRhoSpectrum()
		mirror_ref=mirror_rho
		for i in range(0,len(mirror_rho)):
		    mirror_ref[i][0] = mirror_rho[len(mirror_rho)-1-i][0]/1000.
		    mirror_ref[i][1] = mirror_rho[len(mirror_rho)-1-i][1]/100.
		mirror_ref.append([4,0.9])
		iyaml+='- spectrum: &%s  \n' % 'ref_mirror'
		for i in range(0,len(mirror_ref)-1):
		    iyaml+='  - {wavelength: %15.8e, data: %15.8e }\n' % (float(mirror_ref[i][0]),float(mirror_ref[i][1])) 
		i = len(mirror_ref)-1
		iyaml+='  - {wavelength: %15.8e, data: %15.8e }\n' % (float(mirror_ref[i][0]),float(mirror_ref[i][1])) 
		iyaml+='\n'

	# 
	### Section (2)
	# set the medium types: 
	# air, glass, vacuum, etc. gathering spectral data
	#------------------------------
	#

	#
	# Creation of the sun and atmosphere
	#
	if spectral:
		spectrum = "*solar_spectrum"
	else:
		spectrum = None
	
	iyaml += "- sun: %s\n" % (sun.yaml(spectrum),)

	if medium>1e-99:
		iyaml+='- atmosphere: {extinction: %s}\n'%medium 
		iyaml+='\n'

		   
	# 
	### Section (3)
	# set the materials
	# (gathering media)
	# occultant material, mirror specular material, receiver material, virtual target
	#------------------------------
	#
	# CREATE an occultant material
	r_f = 0. # front
	r_b = 0. # and back reflectivity
	iyaml+='- material: &%s\n' % 'material_black'
	iyaml+='   front:\n'
	iyaml+='     matte: {reflectivity: %6.4f }\n' % r_f    
	iyaml+='   back:\n'
	iyaml+='     matte: {reflectivity: %6.4f }\n' % r_b
	iyaml+='\n'
	#
	# CREATE a specular material
	r_f= rho_refl # front
	r_b = 0.      # and back reflectivity
	iyaml+='- material: &%s\n' % 'material_mirror'
	iyaml+='   front:\n'
	if spectral:
		iyaml+='     mirror: {reflectivity: *%s, slope_error: %15.8e }\n' % ('ref_mirror', slope_error ) 
	else:
		iyaml+='     mirror: {reflectivity: %6.4f, slope_error: %15.8e }\n' % (r_f, slope_error) 

	iyaml+='   back:\n'
	iyaml+='     matte: {reflectivity: %6.4f }\n' % r_b 
	iyaml+='\n'
	#
	# CREATE a material for the target
	r_f = 1.-rec_abs # front
	r_b = 1.-rec_abs # and back reflectivity
	iyaml+='- material: &%s\n' % 'material_target'
	iyaml+='   front:\n'
	iyaml+='     matte: {reflectivity: %6.4f }\n' % r_f    
	iyaml+='   back:\n'
	iyaml+='     matte: {reflectivity: %6.4f }\n' % r_b
	iyaml+='\n'
	#
	# CREATE a virtual material for the calculation of spillage
	iyaml+='- material: &%s\n' % 'material_virtual'
	iyaml+='   virtual:\n'
	iyaml+='\n'


	# 
	### Section (4)
	# set the geometries
	# (gathering shapes and materials)
	# the tower, the receiver, the heliostat
	#------------------------------
	#
	# Tower Geometry
	# (cylindrical shape)
	#
	slices = 10 # slices for the envelop circle
	iyaml+='- geometry: &%s\n' % 'tower_g' 
	iyaml+='  - material: *%s\n' % 'material_black' 
	#iyaml+='    transform: { translation: %s, rotation: %s }\n' % ([0, 0, h_tow*0.5], [0, 90, 0]) 
	iyaml+='    cylinder: {height: %7.3f, radius: %7.3f, slices: %d }\n' % (tower_h, tower_r, slices) 
	iyaml+='\n'
	#
	# Receiver Geometry
	#
	if receiver=='flat':
		geom, rec_entt, rcv = flat_receiver(rec_param, hemisphere)
		iyaml+=geom

	elif receiver=='cylinder':
		geom, rec_entt, rcv = cylindrical_receiver(rec_param, hemisphere)
		iyaml+=geom

	elif receiver=='stl':
		rec_entt, rcv=STL_receiver(rec_param, hemisphere)

	elif receiver=='multi_cavity':
		geom, rec_entt, rcv =multi_cavity_receiver(rec_param, hemisphere)
	#
	# Heliostats Geometry
	#
	if one_heliostat:
		hst_x=np.r_[hst_pos[0]]
		hst_y=np.r_[hst_pos[1]]
		hst_z=np.r_[hst_pos[2]]
		aim_x=np.r_[hst_aims[0]] 
		aim_y=np.r_[hst_aims[1]]
		aim_z=np.r_[hst_aims[2]]
		num_hst=1
		hst_foc=np.r_[hst_foc]
	else:
		hst_x=hst_pos[:,0]
		hst_y=hst_pos[:,1]
		hst_z=hst_pos[:,2]
		aim_x=hst_aims[:,0]
		aim_y=hst_aims[:,1]
		aim_z=hst_aims[:,2]
		num_hst=len(hst_x)
	slices = 4 # slices for the envelop circle
	pts_hst = [ [-hst_w*0.5, -hst_h*0.5], [-hst_w*0.5, hst_h*0.5], [hst_w*0.5, hst_h*0.5], [hst_w*0.5,-hst_h*0.5] ]
	# CREATE a reflective facet (mirror)
	for i in range(0,num_hst):
		name_hst_g = 'hst_g_'+str(i)
		iyaml+='- geometry: &%s\n' % name_hst_g 
		iyaml+='  - material: *%s\n' % 'material_mirror' 
		#iyaml+='    transform: { translation: %s, rotation: %s }\n' % ([hst_x[i], hst_y[i], hst_z[i]], [0, 0, 0]) )
		iyaml+='    parabol: \n'
		iyaml+='      focal: %s\n' % hst_foc[i]
		iyaml+='      clip: \n'  
		iyaml+='      - operation: AND \n'
		iyaml+='        vertices: %s\n' % pts_hst
		iyaml+='      slices: %d\n' % slices  

	# CREATE the pylon "pylon_g" geometry cylindrical shape
	h_pyl = 0.001 # pylon height
	r_pyl = 0.2 # pylon radius
	slices = 4 # slices for the envelop circle
	iyaml+='- geometry: &%s\n' % 'pylon_g' 
	iyaml+='  - material: *%s\n' % 'material_black' 
	iyaml+='    transform: { translation: %s, rotation: %s }\n' % ([0, 0, -h_pyl*3], [0, 90, 0]) 
	iyaml+='    cylinder: {height: %7.3f, radius: %7.3f, slices: %d }\n' % (h_pyl,r_pyl,slices) 
	#   

	# 
	### Section (5)
	# set the templates
	# (programming objects gathering geometries or pivot and geometries)
	#------------------------------
	# CREATE the heliostat templates
	for i in range(0,num_hst):    
		name_hst_t = 'hst_t_'+str(i)
		iyaml+='- template: &%s\n' % name_hst_t 
		name_hst_n = 'hst_'+ str(i)
		iyaml+='    name: %s\n' % name_hst_n 
		iyaml+='    primary: 0\n'   
		iyaml+='    geometry: *pylon_g\n'
		iyaml+='    children: \n' 
		iyaml+='    - name: pivot\n'
		iyaml+='      zx_pivot: {target: {position: %s}} \n' % ([aim_x[i],aim_y[i],aim_z[i]]) 
		iyaml+='      children: \n'
		iyaml+='      - name: reflect_surface\n'
		iyaml+='        primary: 1\n'
		iyaml+='        transform: {rotation: [-90,0,0]} \n'   
		name_hst_g = 'hst_g_'+str(i)
		iyaml+='        geometry: *%s\n' % name_hst_g 

	# 
	### Section (6)
	# set the entities
	# (gather templates to be created and active in the scene)
	#------------------------------
	#
	# receiver entities
	iyaml+=rec_entt
	#
	# tower entities
	iyaml+='\n- entity:\n'
	iyaml+='    name: tower_e\n'
	iyaml+='    primary: 0\n' 
	iyaml+='    transform: { translation: %s, rotation: %s }\n' % ([0, -tower_r, tower_h*0.5], [0, 0, 0]) 
	iyaml+='    geometry: *%s\n' % 'tower_g'    
	#
	# heliostat entities from the template
	for i in range(0,num_hst):
		name_e ='H_'+str(i)
		name_hst_t = 'hst_t_'+str(i)
		iyaml+='\n- entity:\n'
		iyaml+='    name: %s\n' % name_e
		iyaml+='    transform: { translation: %s, rotation: %s }\n' % ([hst_x[i], hst_y[i], hst_z[i]], [0, 0, 0]) 
		iyaml+='    children: [ *%s ]\n' % name_hst_t    

	with open(outfile_yaml,'w') as f:
		f.write(iyaml)

	with open(outfile_recv,'w') as f:
		f.write(rcv) 


def flat_receiver(rec_param, hemisphere='North'):
	"""
	hemisphere : 'North' or 'South' hemisphere of the earth where the field located
		        if North: the field is in the positive y direction
		        if South: the field is in the negtive y direction
		        this will influence:
		         (1) the setting of the receiver tilt angle, 
		             if the front surface always facing to the field is desirable
		         (2) the position of the virtual target
	"""
	rec_w=rec_param[0]
	rec_h=rec_param[1]
	slices=rec_param[2] # it assumes equal number of slices in x and y directions
	x=rec_param[4]
	y=rec_param[5]
	z=rec_param[6]
	tilt=rec_param[7]
	# receiver tilt angle:
	# 0 is vertical
	# the standby posiion of a plane in solstice is normal points to the +z axis
	# rotation anagle, positive is anti-clockwise

	geom=''
	pts=[ [-rec_w*0.5, -rec_h*0.5], [-rec_w*0.5, rec_h*0.5], [rec_w*0.5, rec_h*0.5], [rec_w*0.5,-rec_h*0.5] ]

	geom+='- geometry: &%s\n' % 'target_g'
	geom+='  - material: *%s\n' % 'material_target'
	geom+='    plane: \n'
	geom+='      clip: \n' 
	geom+='      - operation: AND \n'
	geom+='        vertices: %s\n' % pts
	geom+='      slices: %d\n' % slices 
	geom+='\n'

	# CREATE a receiver entity from "target_g" geometry (primary = 0)
	entt=''
	entt+='\n- entity:\n'
	entt+='    name: target_e\n'
	entt+='    primary: 0\n'
	if hemisphere=='North':
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y, z], [-90.-tilt, 0, 0]) 
	else:
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y, z], [90.+tilt, 0, 0]) 
	entt+='    geometry: *%s\n' % 'target_g'

	# CREATE a virtual target entity from "target_g" geometry (primary = 0)
	pts = [ [-rec_w*10., -rec_h*10.], [-rec_w*10., rec_h*10.], [rec_w*10., rec_h*10.], [rec_w*10.,-rec_h*10.] ]
	slices = 4
	entt+='\n- entity:\n'
	entt+='    name: virtual_target_e\n'
	entt+='    primary: 0\n'
	if hemisphere=='North':
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y-5., z], [-90.-tilt, 0, 0])
	else:
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y+5., z], [90.+tilt, 0, 0])
	entt+='    geometry: \n' 
	entt+='      - material: *%s\n' % 'material_virtual' 
	entt+='        plane: \n'
	entt+='          clip: \n'    
	entt+='          - operation: AND \n'
	entt+='            vertices: %s\n' % pts
	entt+='          slices: %d\n' % slices  

	rcv=''
	rcv+='- name: target_e \n' 
	rcv+='  side: %s \n' % 'FRONT_AND_BACK'
	rcv+='  per_primitive: %s \n' % 'INCOMING_AND_ABSORBED'
	rcv+='- name: virtual_target_e \n'
	rcv+='  side: %s \n' % 'FRONT'
	rcv+='  per_primitive: %s \n' % 'INCOMING'

	return geom, entt, rcv



def cylindrical_receiver(rec_param, hemisphere='North'):
	'''
	hemishpere : 'North' or 'South' hemisphere of the earth where the field located
		        if North: the field is in the positive y direction
		        if South: the field is in the negtive y direction
		        this will influence:
		         (1) the setting of the receiver tilt angle, 
		             if the front surface always facing to the field is desirable
		         (2) the position of the virtual target
	'''
	rec_r=rec_param[0]/2.
	rec_h=rec_param[1]
	slices=rec_param[2] # number of elements in the circumferetial direction
	stacks=rec_param[3] # number of elements in the vertical direction
	x=rec_param[4]
	y=rec_param[5]
	z=rec_param[6]

	geom=''
	geom+='- geometry: &%s\n' % 'target_g'
	geom+='  - material: *%s\n' % 'material_target'
	geom+='    cylinder: \n'
	geom+='      height: %s\n'%rec_h 
	geom+='      radius: %s\n'%rec_r 
	geom+='      slices: %d\n' % slices 
	geom+='      stacks: %d\n' % stacks 
	geom+='\n'

	# CREATE a receiver entity from "target_g" geometry (primary = 0)
	entt=''
	entt+='\n- entity:\n'
	entt+='    name: target_e\n'
	entt+='    primary: 0\n'

	entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y, z], [0., 0., 0.]) 

	entt+='    geometry: *%s\n' % 'target_g'

	# CREATE a virtual target entity from "target_g" geometry (primary = 0)
	Vsize=100.
	pts = [ [-rec_h*Vsize, -rec_h*Vsize], [-rec_h*Vsize, rec_h*Vsize], [rec_h*Vsize, rec_h*Vsize], [rec_h*Vsize,-rec_h*Vsize] ]
	slices = 4
	entt+='\n- entity:\n'
	entt+='    name: virtual_target_e\n'
	entt+='    primary: 0\n'

	entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y, z+rec_h/2.+1], [-180., 0, 0])

	entt+='    geometry: \n' 
	entt+='      - material: *%s\n' % 'material_virtual' 
	entt+='        plane: \n'
	entt+='          clip: \n'    
	entt+='          - operation: AND \n'
	entt+='            vertices: %s\n' % pts
	entt+='          slices: %d\n' % slices  

	rcv=''
	rcv+='- name: target_e \n' 
	rcv+='  side: %s \n' % 'FRONT_AND_BACK'
	rcv+='  per_primitive: %s \n' % 'INCOMING_AND_ABSORBED'
	rcv+='- name: virtual_target_e \n'
	rcv+='  side: %s \n' % 'FRONT'
	rcv+='  per_primitive: %s \n' % 'INCOMING'

	return geom, entt, rcv

    

def STL_receiver(rec_param, hemisphere='North'):
	'''
	hemishpere : 'North' or 'South' hemisphere of the earth where the field located
		        if North: the field is in the positive y direction
		        if South: the field is in the negtive y direction
		        this will influence:
		         (1) the setting of the receiver tilt angle, 
		             if the front surface always facing to the field is desirable
		         (2) the position of the virtual target
	'''

	rec_w=rec_param[0].astype(float) # for creating the virtual target
	rec_h=rec_param[1].astype(float)
	stlfile=rec_param[2] # directory of the stl file
	x=rec_param[3].astype(float)
	y=rec_param[4].astype(float)
	z=rec_param[5].astype(float)
	tilt=rec_param[6].astype(float) # need to figure out the initial mesh orientation

	# CREATE a receiver entity from a STL file 
	entt=''
	entt+='\n- entity:\n'
	entt+='    name: STL_receiver_e\n'
	entt+='    primary: 0\n'
	if hemisphere=='North':

		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y, z], [-90.-tilt, 0, 0]) 
	else:
		# if it is the mesh model of the bladed receiver at CSIRO
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y, z], [180.+tilt, 0, 0]) 
	entt+='    geometry:\n'
	entt+='    - material: *material_target\n'
	entt+='      transform: {translation: [0, 0, 0], rotation: [0, 0, 0]}\n'
	entt+="      stl : {path: %s }  \n"%(stlfile)


	# CREATE a virtual target entity from "target_g" geometry (primary = 0)
	pts = [ [-rec_w*10., -rec_h*10.], [-rec_w*10., rec_h*10.], [rec_w*10., rec_h*10.], [rec_w*10.,-rec_h*10.] ]
	slices = 4
	entt+='\n- entity:\n'
	entt+='    name: virtual_target_e\n'
	entt+='    primary: 0\n'
	if hemisphere=='North':
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y-5., z], [-90.-tilt, 0, 0])
	else:
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y+5., z], [90.+tilt, 0, 0])
	entt+='    geometry: \n' 
	entt+='      - material: *%s\n' % 'material_virtual' 
	entt+='        plane: \n'
	entt+='          clip: \n'    
	entt+='          - operation: AND \n'
	entt+='            vertices: %s\n' % pts
	entt+='          slices: %d\n' % slices  

	rcv=''
	rcv+='- name: STL_receiver_e \n' 
	rcv+='  side: %s \n' % 'FRONT_AND_BACK'
	rcv+='  per_primitive: %s \n' % 'INCOMING_AND_ABSORBED'
	rcv+='- name: virtual_target_e \n'
	rcv+='  side: %s \n' % 'FRONT'
	rcv+='  per_primitive: %s \n' % 'INCOMING'

	return entt, rcv

def multi_cavity_receiver(rec_param, hemisphere='North'):
	"""
	hemisphere : 'North' or 'South' hemisphere of the earth where the field located
		        if North: the field is in the positive y direction
		        if South: the field is in the negtive y direction
		        this will influence:
		         (1) the setting of the receiver tilt angle, 
		             if the front surface always facing to the field is desirable
		         (2) the position of the virtual target
	"""
	# x, y, z is the front center of the multi-cavity receiver
	# n_c is the number of cavities
	# rec_w, rec_h is the size of one cavity
	# alpha is the angle between each two cavities 
	# num_cavity is the number of cavities

	rec_w=rec_param[0]
	rec_h=rec_param[1]
	slices=rec_param[2]
	x=rec_param[3]
	y=rec_param[4]
	z=rec_param[5]
	tilt=rec_param[6]
	# receiver tilt angle:
	# 0 is vertical
	# the standby posiion of a plane in solstice is normal points to the +z axis
	# rotation anagle, positive is anti-clockwise

	geom=''
	pts=[ [-rec_w*0.5, -rec_h*0.5], [-rec_w*0.5, rec_h*0.5], [rec_w*0.5, rec_h*0.5], [rec_w*0.5,-rec_h*0.5] ]

	geom+='- geometry: &%s\n' % 'target_g'
	geom+='  - material: *%s\n' % 'material_target'
	geom+='    plane: \n'
	geom+='      clip: \n' 
	geom+='      - operation: AND \n'
	geom+='        vertices: %s\n' % pts
	geom+='      slices: %d\n' % slices 
	geom+='\n'

	# CREATE a receiver entity from "target_g" geometry (primary = 0)
	entt=''
	entt+='\n- entity:\n'
	entt+='    name: target_e\n'
	entt+='    primary: 0\n'
	if hemisphere=='North':
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y, z], [-90.-tilt, 0, 0]) 
	else:
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y, z], [90.+tilt, 0, 0]) 
	entt+='    geometry: *%s\n' % 'target_g'

	# CREATE a virtual target entity from "target_g" geometry (primary = 0)
	pts = [ [-rec_w*10., -rec_h*10.], [-rec_w*10., rec_h*10.], [rec_w*10., rec_h*10.], [rec_w*10.,-rec_h*10.] ]
	slices = 4
	entt+='\n- entity:\n'
	entt+='    name: virtual_target_e\n'
	entt+='    primary: 0\n'
	if hemisphere=='North':
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y-5., z], [-90.-tilt, 0, 0])
	else:
		entt+='    transform: { translation: %s, rotation: %s }\n' % ([x, y+5., z], [90.+tilt, 0, 0])
	entt+='    geometry: \n' 
	entt+='      - material: *%s\n' % 'material_virtual' 
	entt+='        plane: \n'
	entt+='          clip: \n'    
	entt+='          - operation: AND \n'
	entt+='            vertices: %s\n' % pts
	entt+='          slices: %d\n' % slices  

	rcv=''
	rcv+='- name: target_e \n' 
	rcv+='  side: %s \n' % 'FRONT_AND_BACK'
	rcv+='  per_primitive: %s \n' % 'INCOMING_AND_ABSORBED'
	rcv+='- name: virtual_target_e \n'
	rcv+='  side: %s \n' % 'FRONT'
	rcv+='  per_primitive: %s \n' % 'INCOMING'

	return geom, entt, rcv
#------------------------------




#!/usr/bin/env python

import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(name='solsticepy'
	,version='0.3.9'
	,author='Ye Wang'
	,author_email='ye.wang@anu.edu.au'
	,description="Scripts and wrappers for Solstice ray-tracing software"
	,long_description=long_description
	,long_description_content_type="text/markdown"
	,url='https://github.com/anustg/solstice-scripts'
	,packages=setuptools.find_packages()
	,license="GPL v3.0 or later, see LICENSE file"
	,classifiers=[
		"Development Status :: 4 - Beta"
		,"Environment :: Console"
		,"Intended Audience :: Science/Research"
		,"License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)"
		,"Natural Language :: English"
		,"Operating System :: Microsoft :: Windows :: Windows 10"
		,"Operating System :: POSIX :: Linux"
		,"Programming Language :: Python :: 2"
		,"Programming Language :: Python :: 3"
		,"Topic :: Scientific/Engineering :: Physics"
	]
	,install_requires=['scipy','numpy','uncertainties','matplotlib','colorama']
	,python_requires='>=2.7'
)




from __future__ import absolute_import, unicode_literals
from .celery import app
import solsticepy
import os
import numpy as np

@app.task
def add(x, y):
    return x + y


@app.task
def mul(x, y):
    return x * y


@app.task
def xsum(numbers):
    return sum(numbers)


@app.task
def solstice_eta_opt(azimuth,elevation,topdir=None):
	"""
	azimuth=270.   # from East to North, deg
	elevation =78. # 0 is horizontal, deg
	"""

	sun = solsticepy.Sun(dni=1000, sunshape='pillbox', half_angle_deg=0.2664)

	# S3. sun position
	# e.g. summer solstice, solar noon
	latitude=34.   # latitude of the crs plant
	# S4. number of rays for the ray-tracing simulation
	num_rays=2000000
	#
	# the field
	# ==========
	# F1.Layout
	layoutfile='/home/john/solstice-scripts/example/demo_layout.csv'

	# F2. Heliostat
	hst_w=10. # m
	hst_h=10. # m
	rho_refl=0.95 # mirror reflectivity
	slope_error=2.e-3 # radians
	# F3. Tower
	tower_h=0.01 # tower height
	tower_r=0.01 # tower radius
	#
	# the receiver
	# ============
	# R1. shape
	receiver='flat' # 'flat' or 'stl'
	# receiver mesh, for binning the flux distribution
	rec_mesh=100

	# R2. Size
	rec_w=8. # width, m
	rec_h=6. # height, m
	# R3. tilt angle
	tilt=0.  # deg
	# R4. position
	loc_x=0. # m
	loc_y=0. # m
	loc_z=62.# m
	# R5. Abosrptivity
	rec_abs=0.9	

	# extract the heliostat positions from the loaded CSV file
	layout=np.loadtxt(layoutfile, delimiter=',', skiprows=2)
	hst_pos=layout[:,:3]
	hst_foc=layout[:,3] # F2.5
	hst_aims=layout[:,4:] # F4.
	one_heliostat=False

	# define a unique case folder for the user
	snum = 0
	suffix = ""
	while 1:
		import datetime,time
		dt = datetime.datetime.now()
		ds = dt.strftime("%a-%H-%M")
		if topdir is None:
			topdir = os.getcwd()
		casefolder = os.path.join(topdir,'case-%s-%s%s-%d'%(os.path.basename(__file__),ds,suffix,os.getpid()))
		if os.path.exists(casefolder):
		    snum+=1
		    suffix = "-%d"%(snum,)
		    if snum > 200:
		        raise RuntimeError("Some problem with creating casefolder")
		else:
		    # good, we have a new case dir
		    break

	rec_param=np.r_[rec_w, rec_h, rec_mesh, loc_x, loc_y, loc_z, tilt]

	master=solsticepy.Master(casedir=casefolder)
	outfile_yaml = master.in_case('input.yaml')
	outfile_recv = master.in_case('input-rcv.yaml')

	solsticepy.gen_yaml(sun, hst_pos, hst_foc, hst_aims,hst_w, hst_h
		, rho_refl, slope_error, receiver, rec_param, rec_abs
		, outfile_yaml=outfile_yaml, outfile_recv=outfile_recv
		, hemisphere='North', tower_h=tower_h, tower_r=tower_r,  spectral=False
		, medium=0, one_heliostat=one_heliostat)

	# run Solstice using the generate inputs, and run all required post-processing
	eta = master.run(azimuth, elevation, num_rays, rho_refl,sun.dni,gen_vtk=False)

	return (eta.n, eta.s)


